# Copyright 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bridge between :mod:`twisted.python.log` and :mod:`logging`."""

import logging
from functools import partial


class TxLogger(object):

    """Wraps twisted's log object as a ``logging.Logger``."""

    def __init__(self, log):
        """Initialize a ``TxLogger``.

        :param log: A twisted logger (has ``msg`` and ``err`` methods)

        """
        self._log = log

        self.debug = partial(self._msg, logging.DEBUG)
        self.info = partial(self._msg, logging.INFO)
        self.warning = partial(self._msg, logging.WARNING)
        self.exception = self.error = partial(self._msg, logging.ERROR)
        self.log = self._msg

    def _msg(self, lvl, msg, *args, **kwargs):
        try:
            msg = msg % args
        except TypeError:
            pass
        if lvl <= logging.WARNING:
            self._log.msg(msg, **kwargs)
        else:
            self._log.err(None, msg, **kwargs)



# Copyright 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Various utilities for testing txkazoo."""
import mock

from twisted.trial.unittest import TestCase
from txkazoo import TxKazooClient


class TxKazooTestCase(TestCase):

    """Test case mixin for txkazoo tests."""

    def setUp(self):
        """Mock actual KazooClient and deferToThread."""
        self.kazoo_client = mock.patch('kazoo.client.KazooClient').start()
        self.kz_obj = self.kazoo_client.return_value
        self.defer_to_thread = mock.patch(
            'twisted.internet.threads.deferToThread').start()
        self.txkzclient = TxKazooClient(hosts='abc', threads=20)

    def tearDown(self):
        """Stop the patching."""
        mock.patch.stopall()



# Copyright 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the txkazoo equivalent of ``kazoo.recipe.lock``."""
import mock

from txkazoo.recipe.lock import Lock
from twisted.internet import defer
from txkazoo.test.util import TxKazooTestCase


class LockTests(TxKazooTestCase):

    """Tests for `Lock`."""

    def test_method(self):
        """Any method invocation happens in seperate thread."""
        self.defer_to_thread.return_value = defer.succeed(4)
        _lock = mock.Mock()
        lock = Lock(_lock)
        d = lock.acquire(timeout=10)
        self.assertEqual(self.successResultOf(d), 4)
        self.defer_to_thread.assert_called_once_with(_lock.acquire, timeout=10)



# Copyright 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for Twisted-wrapped Kazoo logging.
"""
from logging import INFO
from mock import Mock
from twisted.trial.unittest import SynchronousTestCase
from txkazoo.log import TxLogger


class TxLoggerTests(SynchronousTestCase):

    """Tests for ``TxLogger``."""

    def setUp(self):
        self.log = Mock(spec=['msg', 'err'])
        self.txlog = TxLogger(self.log)

    def test_log(self):
        self.txlog.log(INFO, 'abc', c=2)
        self.log.msg.assert_called_once_with('abc', c=2)

    def test_log_arg(self):
        self.txlog.log(INFO, 'abc %d', 3, c=2)
        self.log.msg.assert_called_once_with('abc 3', c=2)

    def test_log_arg_fail(self):
        self.txlog.log(INFO, 'abc %d', 'xyz', c=2)
        self.log.msg.assert_called_once_with('abc %d', c=2)

    def test_debug(self):
        self.txlog.debug('abc', d=4)
        self.log.msg.assert_called_once_with('abc', d=4)

    def test_info(self):
        self.txlog.info('abc', d=4)
        self.log.msg.assert_called_once_with('abc', d=4)

    def test_warning(self):
        self.txlog.info('abc', d=4)
        self.log.msg.assert_called_once_with('abc', d=4)

    def test_error(self):
        self.txlog.error('abc', d=4)
        self.log.err.assert_called_once_with(None, 'abc', d=4)

    def test_exception(self):
        self.txlog.exception('abc', d=4)
        self.log.err.assert_called_once_with(None, 'abc', d=4)



# Copyright 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from txkazoo.client import TxKazooClient
from txkazoo.test.util import TxKazooTestCase


class TxKazooClientTests(TxKazooTestCase):

    """Tests for `TxKazooClient`."""
    @mock.patch('txkazoo.client.reactor')
    def test_init(self, mock_reactor):
        """__init__ sets up thread size and creates KazooClient."""
        self.txkzclient = TxKazooClient(hosts='abc', arg2='12', threads=20)
        mock_reactor.suggestThreadPoolSize.assert_called_once_with(20)
        self.kazoo_client.assert_called_with(hosts='abc', arg2='12')
        self.assertEqual(self.txkzclient.client, self.kz_obj)

    def test_method(self):
        """Any method is called in seperate thread."""
        d = self.txkzclient.start()
        self.defer_to_thread.assert_called_once_with(
            self.txkzclient.client.start)
        self.assertEqual(d, self.defer_to_thread.return_value)

    def test_property_get(self):
        """Accessing property does not defer to thread.

        It is returned immediately

        """
        s = self.txkzclient.state
        self.assertFalse(self.defer_to_thread.called)
        self.assertEqual(s, self.kazoo_client.return_value.state)



# Copyright 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the txkazoo equivalent of ``kazoo.recipe.partitioner``."""
import mock

from kazoo.recipe.partitioner import PartitionState
from twisted.internet import defer
from txkazoo.test.util import TxKazooTestCase


class SetPartitionerTests(TxKazooTestCase):

    """Tests for txkazoo's ``SetPartitioner``."""

    def test_init(self):
        """Init calls actual SetPartitioner in seperate thread."""
        kz_part = mock.Mock()
        self.defer_to_thread.return_value = defer.succeed(kz_part)
        part = self.txkzclient.SetPartitioner(
            '/path', set(range(2, 5)), time_boundary=20)
        self.defer_to_thread.assert_called_with(
            self.kz_obj.SetPartitioner,
            '/path', set(range(2, 5)), time_boundary=20)
        self.assertEqual(part._partitioner, kz_part)
        self.assertIsNone(part._state)

    def test_state_before_object(self):
        """.state returns ALLOCATING before SetPartitioner object is
        created."""
        self.defer_to_thread.return_value = defer.Deferred()
        partitioner = self.txkzclient.SetPartitioner(
            '/path', set(range(1, 10)))
        self.assertEqual(partitioner.state, PartitionState.ALLOCATING)

    def test_state_after_object(self):
        """.state returns SetPartitioner.state after object is created."""
        kz_part = mock.Mock(state='allocated')
        self.defer_to_thread.return_value = defer.succeed(kz_part)
        partitioner = self.txkzclient.SetPartitioner(
            '/path', set(range(1, 10)))
        self.assertEqual(partitioner.state, 'allocated')

    def test_state_on_error(self):
        """.state returns FAILURE if SetPartitioner creation errored."""
        self.defer_to_thread.return_value = defer.fail(ValueError(2))
        partitioner = self.txkzclient.SetPartitioner(
            '/path', set(range(1, 10)))
        self.assertEqual(partitioner.state, PartitionState.FAILURE)

    def test_state_based_properties(self):
        """accessing other state based properties return True or False
        depending on current value of `state`"""
        attrs = {
            'failed': PartitionState.FAILURE,
            'release': PartitionState.RELEASE,
            'acquired': PartitionState.ACQUIRED,
            'allocating': PartitionState.ALLOCATING
        }
        partitioner = self.txkzclient.SetPartitioner(
            '/path', set(range(1, 10)))
        for attr, value in attrs.items():
            partitioner._state = value
            self.assertTrue(getattr(partitioner, attr))
            partitioner._state = 'else'
            self.assertFalse(getattr(partitioner, attr))

    def test_method_invocation(self):
        """Methods are invoked in seperate thread."""
        # First deferToThread to create SetPartitioner
        part_obj = mock.Mock()
        self.defer_to_thread.return_value = defer.succeed(part_obj)
        partitioner = self.txkzclient.SetPartitioner(
            '/path', set(range(1, 10)))
        # Second deferToThread to call wait_for_acquire
        self.defer_to_thread.return_value = defer.succeed(3)
        d = partitioner.wait_for_acquire(timeout=40)
        self.assertEqual(self.successResultOf(d), 3)
        self.defer_to_thread.assert_called_with(
            part_obj.wait_for_acquire, timeout=40)

    def test_iter(self):
        """__iter__() delegates to actual SetPartitioner.__iter__"""
        part_obj = mock.MagicMock()
        part_obj.__iter__.return_value = [2, 3]
        self.defer_to_thread.return_value = defer.succeed(part_obj)
        partitioner = self.txkzclient.SetPartitioner(
            '/path', set(range(1, 10)))
        self.assertEqual(list(partitioner), [2, 3])



# Copyright 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The txkazoo equivalent of ``kazoo.recipe.lock``."""

from functools import partial
from twisted.internet import threads


class Lock(object):

    """
    Twisted-friendly wrapper for :class:`kazoo.recipe.lock.Lock`.

    All blocking methods of that class are delegated to a thread pool.
    """

    def __init__(self, lock):
        """Initialize the txkazoo Lock.

        :param lock: the blocking kazoo lock being wrapped.
        :type lock: :class:`kazoo.recipe.lock.Lock`

        """
        self._lock = lock

    def __getattr__(self, name):
        """Get a method of the lock and wraps with a thread pool."""
        blocking_method = getattr(self._lock, name)
        return partial(threads.deferToThread, blocking_method)



# Copyright 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test txkazoo test remnants."""
from __future__ import print_function
import sys

from twisted.internet import task, defer
from twisted.python import log
from txkazoo import TxKazooClient


@defer.inlineCallbacks
def partitioning(reactor, client):
    client.add_listener(zk_listener)
    part = client.SetPartitioner(
        '/manitest_partition', set(range(1, 11)), time_boundary=15)
    start = reactor.seconds()
    while True:
        print('current state', client.state)
        if part.failed:
            print('failed, creating new')
            part = client.SetPartitioner(
                '/manitest_partition', set(range(1, 11)))
        if part.release:
            print('part changed. releasing')
            start = reactor.seconds()
            yield part.release_set()
        elif part.acquired:
            print('got part {} in {} seconds'.format(
                list(part), reactor.seconds() - start))
        elif part.allocating:
            print('allocating')
        d = defer.Deferred()
        reactor.callLater(1, d.callback, None)
        yield d


def zk_listener(state):
    print('state change', state)


@defer.inlineCallbacks
def state_changes(reactor, client):
    client.add_listener(zk_listener)
    while True:
        print('state', client.state)
        d = defer.Deferred()
        reactor.callLater(1, d.callback, None)
        yield d


@defer.inlineCallbacks
def locking(reactor, client):
    lock = client.Lock('/locks')
    yield lock.acquire()
    print('got lock')
    d = defer.Deferred()
    reactor.callLater(100, d.callback, None)
    yield d
    yield lock.release()


@defer.inlineCallbacks
def test_via_cli(reactor, hosts):
    log.startLogging(sys.stdout)
    client = TxKazooClient(hosts=hosts, txlog=log)
    yield client.start()
    yield partitioning(reactor, client)
    # yield locking(reactor, client)
    yield client.stop()


if __name__ == '__main__':
    task.react(test_via_cli, sys.argv[1:])



# Copyright 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The Twistified Kazoo client."""
from functools import partial
from txkazoo.log import TxLogger
from kazoo import client
from twisted.internet import reactor, threads
from txkazoo.recipe.lock import Lock
from txkazoo.recipe.partitioner import SetPartitioner


class TxKazooClient(object):

    """Twisted wrapper for `kazoo.client.KazooClient`.

    Runs blocking methods of `kazoo.client.KazooClient` in a different
    thread, returning Deferreds that fire with the method's return value
    or errback with any exception occurred during method execution.

    """

    kz_get_attributes = [
        'handler',
        'retry',
        'state',
        'client_state',
        'client_id',
        'connected'
    ]

    def __init__(self, **kwargs):
        """Initialize `TxKazooClient`.

        Takes same arguments as KazooClient and extra keyword argument
        `threads` that suggests thread pool size to be used

        """
        num_threads = kwargs.pop('threads', 10)
        reactor.suggestThreadPoolSize(num_threads)

        log = kwargs.pop('txlog', None)
        if log:
            kwargs['logger'] = TxLogger(log)

        self.client = client.KazooClient(**kwargs)
        self._internal_listeners = dict()

    def __getattr__(self, name):
        """Get and maybe wrap an attribute from the wrapped client.

        If ``name`` refers to a blocking method, executes the method
        in a thread pool. Otherwise, perform regular attribute access
        on the wrapped client.

        :return: if ``name`` is the name of a blocking method, a
                 :class:`twisted.internet.defer.Deferred` that fires
                 with result of requested method. Otherwise, the
                 wrapped client's attribute with the given name.

        """
        if name in self.kz_get_attributes:
            # Assuming all attributes access are not blocking
            return getattr(self.client, name)

        blocking_method = getattr(self.client, name)
        return partial(threads.deferToThread, blocking_method)

    def add_listener(self, listener):
        """Add the given listener to the wrapped client.

        Even though the event will be coming from the txkazoo thread,
        the listener will be called in the reactor thread. This method
        should only be called from the reactor thread.

        """
        def _listener(state):
            reactor.callFromThread(listener, state)

        self._internal_listeners[listener] = _listener
        return self.client.add_listener(_listener)

    def remove_listener(self, listener):
        """Remove the given listener from the wrapped client."""
        _listener = self._internal_listeners.pop(listener)
        self.client.remove_listener(_listener)

    def _watch_func(self, func, path, watch=None, **kwargs):
        if not watch:
            return threads.deferToThread(func, path, **kwargs)

        def _watch(event):
            # Called from kazoo thread. Replaying in reactor
            reactor.callFromThread(watch, event)

        return threads.deferToThread(func, path, watch=_watch, **kwargs)

    def exists(self, path, watch=None):
        """See py:func:`kazoo.client.KazooClient.exists`."""
        return self._watch_func(self.client.exists, path, watch)

    def exists_async(self, path, watch=None):
        """See py:func:`kazoo.client.KazooClient.exists_async`."""
        return self._watch_func(self.client.exists_async, path, watch)

    def get(self, path, watch=None):
        """See py:func:`kazoo.client.KazooClient.get`."""
        return self._watch_func(self.client.get, path, watch)

    def get_async(self, path, watch=None):
        """See py:func:`kazoo.client.KazooClient.get_async`."""
        return self._watch_func(self.client.get_async, path, watch)

    def get_children(self, path, watch=None, include_data=False):
        """See py:func:`kazoo.client.KazooClient.get_children`."""
        return self._watch_func(self.client.get_children,
                                path, watch, include_data=include_data)

    def get_children_async(self, path, watch=None, include_data=False):
        """See py:func:`kazoo.client.KazooClient.get_children_async`."""
        return self._watch_func(self.client.get_children_async,
                                path, watch, include_data=include_data)

    def Lock(self, path, identifier=None):
        """Return a wrapped ``Lock`` for this client."""
        return Lock(self.client.Lock(path, identifier))

    def SetPartitioner(self, path, set, **kwargs):
        """Return a wrapped ``SetPartitioner`` for this client."""
        return SetPartitioner(self.client, path, set, **kwargs)



# Copyright 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Twisted binding for Kazoo.

This doesn't really reimplement Kazoo with Twisted. Instead it delegates
all blocking calls to seperate thread and returns result as Deferred.
This allows usage of using Kazoo in Twisted reactor thread without
blocking it.

"""
from txkazoo._version import __version__
version = __version__

from txkazoo.client import TxKazooClient
from txkazoo.log import TxLogger
from txkazoo.recipe.lock import Lock
from txkazoo.recipe.partitioner import SetPartitioner
__all__ = ("TxKazooClient", "TxLogger", "Lock", "SetPartitioner")



from os.path import dirname, join
from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand
from sys import exit

package_name = "txkazoo"

def read(path):
    with open(join(dirname(__file__), path)) as f:
        return f.read()

import re
version_line = read("{0}/_version.py".format(package_name))
match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]$", version_line, re.M)
version_string = match.group(1)

dependencies = map(str.split, read("requirements.txt").split())

class Tox(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import tox
        exit(tox.cmdline([]))

setup(name=package_name,
      version=version_string,
      description='Twisted binding for Kazoo',
      long_description=read("README.md"),
      url="https://github.com/rackerlabs/txkazoo",

      author='Manish Tomar',
      author_email='manish.tomar@rackspace.com',
      maintainer='Manish Tomar',
      maintainer_email='manish.tomar@rackspace.com',

      license='Apache 2.0',
      keywords="twisted kazoo zookeeper distributed",
      classifiers=[
          "Development Status :: 3 - Alpha",
          "Framework :: Twisted",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: Apache Software License",
          "Programming Language :: Python :: 2 :: Only",
          "Programming Language :: Python :: 2.7",
          "Topic :: System :: Distributed Computing"
      ],

      packages=find_packages(),
      install_requires=dependencies,
      test_suite="{0}.test".format(package_name),
      cmdclass={'test': Tox},

      zip_safe=True)



# Copyright 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The txkazoo equivalent of ``kazoo.recipe.partitioner``."""

from functools import partial
from kazoo.recipe.partitioner import PartitionState
from twisted.internet import threads


class SetPartitioner(object):

    """Wrapper for :class:`kazoo.recipe.partitioner.SetPartitioner`.

    This is a Twisted-friendly wrapped based on a thread pool.
    """

    def __init__(self, client, path, set, **kwargs):
        """Initialize SetPartitioner.

        After the ``client`` argument, takes same arguments as
        :class:`kazoo.recipe.partitioner.SetPartitioner.__init__`.

        :param client: blocking kazoo client
        :type client: :class:`kazoo.client.KazooClient`

        """
        self._partitioner = None
        self._state = PartitionState.ALLOCATING
        d = threads.deferToThread(client.SetPartitioner, path, set, **kwargs)
        d.addCallback(self._initialized)
        d.addErrback(self._errored)

    def _initialized(self, partitioner):
        """Store the partitioner and reset the internal state.

        Now that we successfully got an actual
        :class:`kazoo.recipe.partitioner.SetPartitioner` object, we
        store it and reset our internal ``_state`` to ``None``,
        causing the ``state`` property to defer to the partitioner's
        state.

        """
        self._partitioner = partitioner
        self._state = None

    def _errored(self, failure):
        """Remember that we failed to initialize.

        This means we couldn't get a
        :class:`kazoo.recipe.partitioner.SetPartitioner`: most likely
        a session expired or a network error occurred. The internal
        state is set to ``PartitionState.FAILURE``.

        """
        self._state = PartitionState.FAILURE

    @property
    def state(self):
        """The current state of this partitioner.

        If we are still initializing, or we've failed to initialize,
        this will be this object's internal state. Otherwise, defers
        to the partitioner's state.

        """
        return self._state or self._partitioner.state

    @property
    def allocating(self):
        """Check if the partitioner is still allocating.

        This means either we're still getting a partitioner, or the
        partitioner itself is still allocating (see
        :py:func:`kazoo.recipe.partitioner.Partitioner.allocating`).

        """
        return self.state == PartitionState.ALLOCATING

    @property
    def failed(self):
        """Check if the partitioner has failed.

        This means we've either failed to get a partitioner, or the
        partitioner itself has failed to partition the set (see
        :py:func:`kazoo.recipe.partitioner.Partitioner.failed`).

        """
        return self.state == PartitionState.FAILURE

    @property
    def release(self):
        """Check if the set needs to be repartitioned.

        See :py:func:`kazoo.recipe.partitioner.Partitioner.released`.

        """
        return self.state == PartitionState.RELEASE

    @property
    def acquired(self):
        """Check if the set partitioning has been acquired.

        See :py:func:`kazoo.recipe.partitioner.Partitioner.acquired`.

        """
        return self.state == PartitionState.ACQUIRED

    def __getattr__(self, name):
        """Get a method of the partitioner and wraps with a thread pool."""
        blocking_method = getattr(self._partitioner, name)
        return partial(threads.deferToThread, blocking_method)

    def __iter__(self):
        """Iterate over the wrapped partitioner."""
        return iter(self._partitioner)
